﻿/*
 *
 *
 *    AutoCapV1.cs
 *
 *    Interface to custom AutoCap version 1.0 motor controller.
 *
 *    License: GNU General Public License Version 3.0.
 *    
 *    Copyright (C) 2014-2018 by Matthew K. Roberts, KK5JY.
 *    All rights reserved.
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *    
 *    This program is distributed in the hope that it will be useful, but
 *    WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    General Public License for more details.
 *    
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see: http://www.gnu.org/licenses/
 *    
 *
 */

#if DEBUG
// define this to make the debug output chat about the motor command progress
#define VERBOSE_DEBUG
#endif

using System;
using System.IO.Ports;
using System.Diagnostics;
using System.Threading;
using System.Net;

using KK5JY.IO;

//
//  TODO:
//    1. Parse and store current readings from the microcontroller's A/D.
//

namespace KK5JY.AutoCap {
	/// <summary>
	/// A motor connected to AutoCap external controller with version 1.0 firmware.
	/// </summary>
	public class AutoCapMotorV1 : ITimedMotor, IBrakedMotor, IStepperMotor, IStart, IQueryVersion, IDisposable {
		/// <summary>
		/// The line ending to send with each command.
		/// </summary>
		public const string LineEnding = "\r\n";

		/// <summary>
		/// The amount of time to wait for a command response before allowing
		/// the next command to be sent without it.
		/// </summary>
		public readonly TimeSpan CommandTimeout = TimeSpan.FromSeconds(1.0);

		#region Public Types and Properties
		/// <summary>
		/// Indicates that the motor should be rotated opposite
		/// to the direction commanded.
		/// </summary>
		public bool MotorReversed { get; set; }

		/// <summary>
		/// The output port number, normally zero or one.
		/// </summary>
		public ushort OutputPort { get; set; }

		/// <summary>
		/// The version string read from the controller.
		/// </summary>
		public string VersionString { get; private set; }

		/// <summary>
		/// The current brake state reported by the controller
		/// </summary>
		public bool Brake { get; private set; }

		/// <summary>
		/// Return the most current position report.
		/// </summary>
		public long Position {
			get {
				return Interlocked.Read(ref mPosition);
			}
		}

		/// <summary>
		/// The auto-poll interval, measured in seconds.
		/// </summary>
		public decimal AutoPollInterval {
			get { return m_AutoPollInterval; }
			set {
				m_AutoPollInterval = value;
				if (!ReferenceEquals(m_Timer, null)) {
					SetTimer();
				}
			}
		}

		/// <summary>
		/// Set to true to automatically poll the position of a stepper motor.s
		/// </summary>
		public bool AutoPollPosition {
			get {
				return !ReferenceEquals(m_Timer, null);
			}
			set {
				if (value) {
					if (!ReferenceEquals(m_Timer, null)) return;
					SetTimer();
				} else {
					if (ReferenceEquals(m_Timer, null)) return;
					m_Timer.Dispose();
					m_Timer = null;
				}
			}
		}
		#endregion

		#region Private Fields
		/// <summary>
		/// The serial port object.
		/// </summary>
		private IDeviceHelper m_Peer;

		/// <summary>
		/// The RX buffer.
		/// </summary>
		private string m_RxBuffer;

		/// <summary>
		/// The last outstanding command string.
		/// </summary>
		private string m_LastCommand;

		/// <summary>
		/// The time of the last command.
		/// </summary>
		private DateTimeOffset m_LastCommandTime;

		/// <summary>
		/// The most recent position report.
		/// </summary>
		private long mPosition;

		/// <summary>
		/// Used for auto-polling.
		/// </summary>
		private Timer m_Timer;

		/// <summary>
		/// The position polling interval.
		/// </summary>
		private decimal m_AutoPollInterval;
		#endregion

		#region Constructors
		/// <summary>
		/// Constructor.
		/// </summary>
		/// <param name="portName">The serial port.</param>
		/// <param name="speed">The port speed.</param>
		public AutoCapMotorV1(string portName, int speed = 9600) {
			var port = new SerialPort(portName);
			m_Peer = new SerialPortHelper(port);
			m_Peer.DataReceived += ProcessRxBuffer;
			AutoPollInterval = 1m; // 1 sec

			port.BaudRate = speed;
			port.Parity = Parity.None;
			port.StopBits = StopBits.One;
			port.DataBits = 8;
			port.Handshake = Handshake.None;
		}

		/// <summary>
		/// Constructor.
		/// </summary>
		/// <param name="ep">The remote IP endpoint address.</param>
		public AutoCapMotorV1(IPEndPoint ep) {
			m_Peer = new SocketHelper(ep);
			m_Peer.DataReceived += ProcessRxBuffer;
			AutoPollInterval = 1m; // 1 sec
		}
		#endregion

		#region Public Methods
		/// <summary>
		/// Shut down the motor.
		/// </summary>
		public void Dispose() {
			try { m_Peer.Dispose(); } catch { /* nop */ }
			var timer = m_Timer;
			if (!ReferenceEquals(timer, null)) {
				timer.Dispose();
				m_Timer = null;
			}
		}

		/// <summary>
		/// Start the controller.
		/// </summary>
		public void Start() {
			m_Peer.Start();
		}

		/// <summary>
		/// Move the motor.
		/// </summary>
		/// <param name="speed">The speed (in this case, just the direction).</param>
		public bool Move(decimal speed = 0) {
			// clamp the speed
			if (speed > 1m) speed = 1m;
			if (speed < -1m) speed = -1m;

			// invert?
			if (MotorReversed)
				speed *= -1m;

			char dir = 'U';
			if (speed < 0) {
				dir = 'D';
			}
			var port = OutputPort.ToString()[0];
			int effort = Convert.ToInt32(Math.Abs(speed) * 255m);

			// send the command
			var cmd = String.Format("M{0}{1}{2:X2}{3}", dir, port, effort, LineEnding);
			return SendMessage(cmd);
		}

		/// <summary>
		/// Generate a movement pulse.
		/// </summary>
		/// <param name="speed">The normalized effort (negative is reversed).</param>
		/// <param name="duration">The duration, in seconds.</param>
		public bool Pulse(decimal speed, decimal duration) {
			// clamp the speed
			if (speed > 1m) speed = 1m;
			if (speed < -1m) speed = -1m;

			// invert?
			if (MotorReversed)
				speed *= -1m;

			char dir = 'U';
			if (speed < 0) {
				dir = 'D';
			}
			var port = OutputPort.ToString()[0];
			int effort = Convert.ToInt32(Math.Abs(speed) * 255m);
			int width = Convert.ToInt32(duration * 1000m);

			// send the command
			var cmd = String.Format("P{0}{1}{2:X4}{3:X2}{4}", dir, port, width, effort, LineEnding);
			return SendMessage(cmd);
		}

		/// <summary>
		/// Step forward or backward (negative).
		/// </summary>
		/// <param name="speed">The normalized stepping rate (negative is reversed).</param>
		/// <param name="steps">The number of steps.</param>
		/// <returns>True if the command was successfully sent.</returns>
		public bool Step(decimal speed, ulong steps) {
			// clamp the speed
			if (speed > 1m) speed = 1m;
			if (speed < -1m) speed = -1m;

			// if we can't encode this in a single command, give up
			if (steps >= UInt16.MaxValue)
				return false;

			// invert?
			if (MotorReversed)
				speed *= -1m;

			char dir = 'U';
			if (speed < 0) {
				dir = 'D';
			}
			var port = OutputPort.ToString()[0];
			int effort = Convert.ToInt32(Math.Abs(speed) * 255m);
			ushort shortSteps = Convert.ToUInt16(steps);

			// send the command
			var cmd = String.Format("T{0}{1}{2:X4}{3:X2}{4}", dir, port, shortSteps, effort, LineEnding);
			return SendMessage(cmd);
		}

		/// <summary>
		/// Make sure the motor is stopped.
		/// </summary>
		public bool SetSafeState() {
			return SendMessage("Z" + LineEnding, true);
		}

		/// <summary>
		/// Enable/disable braking.
		/// </summary>
		/// <param name="enabled">Enable (true) or disable (false).</param>
		public bool SetBraking(bool enabled) {
			var port = OutputPort.ToString()[0];
			var enb = enabled ? '1' : '0';

			// send the command
			var cmd = String.Format("B{0}{1}{2}", port, enb, LineEnding);
			return SendMessage(cmd);
		}

		/// <summary>
		/// Enable or disable analog current readbacks.
		/// </summary>
		/// <param name="enabled">Enable (true) or disable (false).</param>
		public bool SetStats(bool enabled) {
			var enb = enabled ? '1' : '0';

			// send the command
			var cmd = String.Format("S{0}{1}", enb, LineEnding);
			return SendMessage(cmd);
		}

		/// <summary>
		/// Query the firmware version number.
		/// </summary>
		public bool QueryVersion() {
			// send the command
			var cmd = String.Format("I{0}", LineEnding);
			return SendMessage(cmd, true);
		}

		/// <summary>
		/// Reset the stepper's 'odometer' position to zero.
		/// </summary>
		/// <returns>True iff successful.</returns>
		public bool ResetPosition() {
			// send the command
			var port = OutputPort.ToString()[0];
			var cmd = String.Format("R{0}{1}", port, LineEnding);
			return SendMessage(cmd, true);
		}

		/// <summary>
		/// Get the current stepper position.
		/// </summary>
		/// <returns>True iff successful.</returns>
		public bool QueryPosition() {
			// send the command
			var port = OutputPort.ToString()[0];
			var cmd = String.Format("X{0}{1}", port, LineEnding);
			return SendMessage(cmd, true);
		}

		/// <summary>
		/// Go to a specific stepper position.
		/// </summary>
		public bool Seek(long target) {
			// send the command
			var port = OutputPort.ToString()[0];
			var cmd = String.Format("X{0}{1}{2}", port, target, LineEnding);
			return SendMessage(cmd, true);
		}
		#endregion

		#region Utilities
		private bool SendMessage(string cmd, bool force = false) {
			var now = DateTimeOffset.UtcNow;
			if (force || String.IsNullOrEmpty(m_LastCommand) || (now - m_LastCommandTime) > CommandTimeout) {
				if (!force) {
					m_LastCommand = cmd.Trim();
					m_LastCommandTime = now;
				}
				try {
#if DEBUG
					Debug.WriteLine("DEBUG: Send: " + cmd);
#endif
					m_Peer.Write(cmd);
					return true;
				} catch {
					// nop - fall through
				}
			}
			return false;
		}

		private void SetTimer() {
			if (!ReferenceEquals(m_Timer, null)) {
				m_Timer.Dispose();
			}
			m_Timer = new Timer(
				TimerTick,
				null,
				Convert.ToInt64(AutoPollInterval * 1000m),
				Convert.ToInt64(AutoPollInterval * 1000m)
			);
		}
		#endregion

		#region Event Handlers
		/// <summary>
		/// Raised when a position change is reported.
		/// </summary>
		public event EventHandler<PositionEventArgs> PositionUpdate;

		/// <summary>
		/// Raised when data received from the controller.
		/// </summary>
		private void ProcessRxBuffer(object sender, ReceivedDataEventArgs e) {
			m_RxBuffer += e.Data;

			int idx = -1;
			do {
				int len = 2;
				idx = m_RxBuffer.IndexOf(LineEnding);
				//if (idx < 0) {
				//    len = 1;
				//    idx = m_RxBuffer.IndexOf('\r');
				//}
				if (idx < 0) {
					len = 1;
					idx = m_RxBuffer.IndexOf('\n');
				}
				if (idx >= 0) {
					var data = m_RxBuffer.Substring(0, idx + len);
					m_RxBuffer = m_RxBuffer.Substring(idx + len);
					
					// slice off leading partial messages
					int idx2 = data.LastIndexOf('#');
					data = data.Substring(idx2);
#if DEBUG
					Debug.WriteLine("DEBUG: Motor Data: " + data.Trim());
#endif
					if (data.StartsWith("#error,", StringComparison.InvariantCultureIgnoreCase)) {
						//
						//  TODO: store this somewhere, raise an event, etc.
						//
					} else if (data.StartsWith("#info,", StringComparison.InvariantCultureIgnoreCase)) {
						VersionString = data.Substring(6).Trim();
#if DEBUG
						Console.WriteLine("DEBUG: Motor = " + VersionString);
#endif
					} else if (data.StartsWith("#OK,", StringComparison.InvariantCultureIgnoreCase)) {
						data = data.Substring(4).Trim();
#if DEBUG
						Debug.WriteLine("DEBUG: ACK! (" + data.Trim() + ":" + (m_LastCommand ?? "") + ")");
#endif
						if (String.Equals(data, m_LastCommand)) {
							m_LastCommand = null;
						}
						//
						//  process specific command responses
						//
						if (data.Length >= 3) {
							switch (Char.ToLower(data[0])) {
								// BRAKE:
								case 'b': {
									int port;
									if (Int32.TryParse(data.Substring(1, 1), out port)) {
										if (port != OutputPort) break;
										Brake = (data[2] == '1');
									}
								} break;
							}
						}
					} else if (data.StartsWith("#x", StringComparison.InvariantCultureIgnoreCase)) {
						int port;
						if (Int32.TryParse(data.Substring(2, 1), out port)) {
							if (port != OutputPort) break;
							var parts = data.Split(',');
							if (parts.Length != 2) break;
							long pos;
							if (Int64.TryParse(parts[1], out pos)) {
								Interlocked.Exchange(ref mPosition, pos);
							}
							if (PositionUpdate != null) {
								PositionUpdate(this, new PositionEventArgs(pos));
							}
						}
					}

					//
					//  TODO: act on other information
					//
				}
			} while (idx >= 0);
		}

		/// <summary>
		/// Autopolling timer.
		/// </summary>
		private void TimerTick(object ar) {
			this.QueryPosition();
		}
		#endregion
	}
}
