﻿/*
 *
 *
 *    Core.cs
 *
 *    Tuner core.
 *
 *    License: GNU General Public License Version 3.0.
 *    
 *    Copyright (C) 2014-2018 by Matthew K. Roberts, KK5JY.
 *    All rights reserved.
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *    
 *    This program is distributed in the hope that it will be useful, but
 *    WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    General Public License for more details.
 *    
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see: http://www.gnu.org/licenses/
 *    
 *
 */

#if DEBUG
#define VERBOSE_DEBUG
#endif

using System;
using System.Collections.Generic;
using System.IO.Ports;
using System.Net;
using System.Threading;
using System.Diagnostics;

using KK5JY.IO;

namespace KK5JY.AutoCap {
	/// <summary>
	/// The core of the autotuner.
	/// </summary>
	public class Core : IDisposable {
		/// <summary>
		/// Loop-to-loop delay when no tuning going on
		/// </summary>
		public const int IdlePollDelay = 100; // msec

		/// <summary>
		/// Loop-to-loop delay when waiting for a timed motor pulse to complete
		/// </summary>
		public const int TimedSpinDelay = 25; // msec

		#region Private Types & Fields
#if false
		/// <summary>
		/// Delegate for pulsing the motor.
		/// </summary>
		/// <param name="pulse">The details</param>
		private delegate void PulseDelegate(PulseNode pulse);
#endif
		/// <summary>
		/// User pulse details.
		/// </summary>
		private sealed class PulseNode {
			/// <summary>
			/// The normalized effort for the pulse in [ -1.0 , 1.0 ]
			/// </summary>
			public readonly decimal Effort;

			/// <summary>
			/// The duration of the pulse, measured in seconds.
			/// </summary>
			public readonly decimal Size;

			/// <summary>
			/// Set to true of the Duration is a step count.
			/// </summary>
			public readonly bool IsStep;

			/// <summary>
			/// Create a new pulse object.
			/// </summary>
			/// <param name="effort">The normalized effort.</param>
			/// <param name="size">The duration.</param>
			/// <param name="isStep">True if command is step; otherwise command is pulse.</param>
			public PulseNode(decimal effort, decimal size, bool isStep) {
				Effort = effort;
				Size = size;
				IsStep = isStep;
			}

			/// <summary>
			/// Copy constructor.
			/// </summary>
			/// <param name="n">The node to copy.</param>
			public PulseNode(PulseNode n) {
				Effort = n.Effort;
				Size = n.Size;
				IsStep = n.IsStep;
			}
		}

		/// <summary>
		/// Current user pulse.
		/// </summary>
		private PulseNode m_UserPulse;

		/// The sensor.
		/// </summary>
		private ISensor m_Sensor;

		/// <summary>
		/// The motor control.
		/// </summary>
		private IMotor m_Motor;

		/// <summary>
		/// The radio.
		/// </summary>
		private IRadio m_Radio;

		/// <summary>
		/// The worker thread.
		/// </summary>
		private Thread m_AlgoThread;

		/// <summary>
		/// The most recent SWR reading.
		/// </summary>
		private SensorData m_Data;

		/// <summary>
		/// A frequency hinting counter.
		/// </summary>
		private BipolarCounter m_FreqHintCounter;

		/// <summary>
		/// Version number for the current sensor data.
		/// </summary>
		private ulong m_SensorVersion;

		/// <summary>
		/// Thread exit flag.
		/// </summary>
		private bool m_Quit;

		/// <summary>
		/// Stores user movement commands.
		/// </summary>
		private decimal? m_UserMoveEffort;

#if DEBUG
		/// <summary>
		/// Tracks the sensor input rate.
		/// </summary>
		private DateTimeOffset m_LastSensorRateUpdate;
#endif
		#endregion

		#region Public Properties
		/// <summary>
		/// The configuration object.
		/// </summary>
		public CoreConfig Config { get; private set; }

		/// <summary>
		/// Get the latest data block.
		/// </summary>
		public SensorData Data { get { return m_Data; } }

		/// <summary>
		/// Return the most recent motor effort.
		/// </summary>
		public decimal MotorEffort { get { return m_CurrentAction; } }

		/// <summary>
		/// The most recent frequency read from the radio.
		/// </summary>
		public decimal Frequency { get { return m_LastReadFreq; } }

		/// <summary>
		/// <summary>
		/// The most recent tuned frequency; zero = untuned and/or unavailable.
		/// </summary>
		public decimal TunedFrequency {
			get { return m_LastTunedFreq; }
		}

		/// <summary>
		/// Returns the motor version number, if available.
		/// </summary>
		public string MotorVersion {
			get {
				var reports = m_Motor as IReportsVersion;
				if (!ReferenceEquals(reports, null)) {
					return reports.VersionString;
				} else {
					return null;
				}
			}
		}

		/// <summary>
		/// Returns the sensor version number, if available.
		/// </summary>
		public string SensorVersion {
			get {
				var reports = m_Sensor as IReportsVersion;
				if (!ReferenceEquals(reports, null)) {
					return reports.VersionString;
				} else {
					return null;
				}
			}
		}

		/// <summary>
		/// Returns the radio version number, if available.
		/// </summary>
		public string RadioVersion {
			get {
				var reports = m_Radio as IReportsVersion;
				if (!ReferenceEquals(reports, null)) {
					return reports.VersionString;
				} else {
					return null;
				}
			}
		}

		/// <summary>
		/// Return the current frequency hint.
		/// </summary>
		public int FrequencyHint {
			get { return Convert.ToInt32(m_FreqHintCounter.Value); }
			set { m_FreqHintCounter.Reset(value); }
		}

		/// <summary>
		/// Automatic tuning actions only happen when this property
		/// is set to true.  When set to false, measurement and other
		/// housekeeping still occurs, but the state machine is not
		/// allowed to transition into any of the motor command states.
		/// </summary>
		public bool Automatic { get; set; }
		#endregion

		#region Public Methods
		/// <summary>
		/// Default constructor.
		/// </summary>
		public Core() {
			Config = new CoreConfig();
			m_FreqHintCounter = new BipolarCounter();
			m_SensorVersion = 0ul;
			Automatic = true;
		}

		/// <summary>
		/// Stops the core if it is running.
		/// </summary>
		public void Dispose() {
			Stop();
		}

		/// <summary>
		/// Start the core.
		/// </summary>
		public void Start() {
			// clean up old objects
			foreach (var id in new[] { m_Motor as IDisposable, m_Sensor as IDisposable, m_Radio as IDisposable }) {
				if (!ReferenceEquals(id, null)) {
					try { id.Dispose(); } catch { /* nop */ }
				}
			}
			m_Motor = null;
			m_Sensor = null;
			m_Radio = null;

			// use this for address parsing
			IPEndPoint ep;

			// construct SENSOR object
			try {
				switch (Config.SensorType) {
					case SensorTypes.LP100A:
						if (PortUtils.TryParse(Config.SensorPort, out ep)) {
							m_Sensor = new LP100A(ep);
						} else {
                            if (!PortUtils.PortExists(Config.SensorPort)) {
								throw new ApplicationException("Invalid serial device: " + Config.SensorPort);
							}
							m_Sensor = new LP100A(Config.SensorPort);
						}
						break;
					case SensorTypes.AutoCapV1:
                        if (PortUtils.TryParse(Config.SensorPort, out ep)) {
							m_Sensor = new AutoCapSensorV1(ep);
						} else {
                            if (!PortUtils.PortExists(Config.SensorPort)) {
								throw new ApplicationException("Invalid serial device: " + Config.SensorPort);
							}
							m_Sensor = new AutoCapSensorV1(Config.SensorPort);
						}
						break;
					case SensorTypes.Sim:
						m_Sensor = TestHardware.Instance;
						break;
				}
				m_Sensor.SensorData += SensorDataReceived;
			} catch (Exception ex) {
				throw new ApplicationException("Unable to create " + Config.SensorType.ToString() + " sensor object: " + ex.Message);
			}

			// construct MOTOR object
			try {
				switch (Config.MotorType) {
					case MotorTypes.RS232:
                        if (!PortUtils.PortExists(Config.MotorControlPort)) {
							throw new ApplicationException("Invalid serial device: " + Config.MotorControlPort);
						}
						m_Motor = new RS232Motor(Config.MotorControlPort) {
							WireConfig =
								Config.MotorRelayRolesReversed ?
								RS232Motor.WireConfigOptions.RtsPowerDtrPolarity :
								RS232Motor.WireConfigOptions.DtrPowerRtsPolarity
						};
						break;
					case MotorTypes.RS232Bridge:
                        if (!PortUtils.PortExists(Config.MotorControlPort)) {
							throw new ApplicationException("Invalid serial device: " + Config.MotorControlPort);
						}
						m_Motor = new RS232MotorH(Config.MotorControlPort);
						break;
					case MotorTypes.AutoCapV1:
                        if (PortUtils.TryParse(Config.MotorControlPort, out ep)) {
							m_Motor = new AutoCapMotorV1(ep) {
								OutputPort = Convert.ToUInt16(Config.MotorOutputPort)
							};
						} else {
                            if (!PortUtils.PortExists(Config.MotorControlPort)) {
								throw new ApplicationException("Invalid serial device: " + Config.MotorControlPort);
							}
							m_Motor = new AutoCapMotorV1(Config.MotorControlPort) {
								OutputPort = Convert.ToUInt16(Config.MotorOutputPort)
							};
						}
						break;
					case MotorTypes.Sim:
						m_Motor = TestHardware.Instance;
						break;
				}
				m_Motor.MotorReversed = Config.MotorPolarityReversed;
			} catch (Exception ex) {
				throw new ApplicationException("Unable to create " + Config.MotorType.ToString() + " motor object: " + ex.Message);
			}

			// construct radio object
			try {
				switch (Config.RadioType) {
					case RadioTypes.Sim:
						m_Radio = TestHardware.Instance;
						break;
					case RadioTypes.HRD:
						m_Radio = new RadioHRD(Config.RadioHost); // TODO: bring the TCP port over, too
						break;
					case RadioTypes.FLDigi:
						m_Radio = new RadioFlDigi(Config.RadioHost); // TODO: bring the TCP port over, too
						break;
					default:
						m_Radio = new NullRadio(); // TODO: should be real radio i/f type
						break;
				}
			} catch {
				throw new ApplicationException("Unable to create " + Config.RadioType.ToString() + " radio object (" + (Config.RadioHost ?? "none") + ")");
			}
#if DEBUG
			Debug.Assert(!ReferenceEquals(m_Sensor, null));
			Debug.Assert(!ReferenceEquals(m_Motor, null));
			Debug.Assert(!ReferenceEquals(m_Radio, null));
			Debug.Assert(Config.SensorEpsilon > 0m);
			m_LastSensorRateUpdate = DateTimeOffset.UtcNow;

			Debug.WriteLine("Motor Type: " + Config.MotorType.ToString());
			if (Config.MotorType != MotorTypes.Sim) {
				Debug.WriteLine("Motor Port: " + Config.MotorControlPort);
			}
			Debug.WriteLine("Sensor Type: " + Config.SensorType.ToString());
			if (Config.MotorType != MotorTypes.Sim) {
				Debug.WriteLine("Sensor Port: " + Config.SensorPort);
			}
			Debug.WriteLine("Sensor Epsilon: " + Config.SensorEpsilon.ToString());
#endif
			// perform pre-thread-start initialization
			m_Data = null;

			// start the worker thread
			m_Quit = false;
			if (m_AlgoThread != null) return;
			m_AlgoThread = new Thread(PulseAlgorithmThread);
			m_AlgoThread.Name = "KK5JY.AutoCap.Core.SimpleAlgorithm";
			m_AlgoThread.Start();
		}

		/// <summary>
		/// Stops the core.
		/// </summary>
		public void Stop() {
			m_Quit = true;
			var thread = m_AlgoThread;
			if (thread != null) {
				thread.Join();
				m_AlgoThread = null;
			}

			// clean up old objects
			foreach (var id in new[] { m_Motor as IDisposable, m_Sensor as IDisposable, m_Radio as IDisposable }) {
				if (!ReferenceEquals(id, null)) {
					try { id.Dispose(); } catch { /* nop */ }
				}
			}
			m_Motor = null;
			m_Sensor = null;
			m_Radio = null;

			// clean up some other vars, too
			m_LastTunedFreq = 0m;
			m_LastReadFreq = 0m;
			m_LastAction = 0m;
			m_CurrentAction = 0m;
			m_LastActionSWR = 0m;
			m_UserPulse = null;
			m_Data = null;
		}

		/// <summary>
		/// Move the motor.
		/// </summary>
		/// <param name="effort">The normalized effort, including direction/sign.</param>
		public void Move(decimal effort) {
            // if the user tries to manually move the motor against the current automatic tuning action, reverse the next automatic action
            if (m_LastAction * effort < 0)
            {
                m_LastAction = -m_LastAction;
            }

            // just store it, the main loop will make it work
			m_UserMoveEffort = effort;
		}

		/// <summary>
		/// Pulse the motor.
		/// </summary>
		/// <param name="effort">The normalized effort, including direction/sign.</param>
		/// <param name="seconds">The duration of the pulse in seconds.</param>
		public void Pulse(decimal effort, decimal seconds) {
            // if the user tries to manually move the motor against the current automatic tuning action, reverse the next automatic action
            if (m_LastAction * effort < 0)
            {
                m_LastAction = -m_LastAction;
            }

            // just store it, the main loop will make it work
			m_UserPulse = new PulseNode(effort, seconds, false);
		}

		/// <summary>
		/// Step the motor
		/// </summary>
		/// <param name="effort">The normalized effort, including direction/sign.</param>
		/// <param name="count">The number of steps.</param>
		public void Step(decimal effort, decimal count) {
            // if the user tries to manually move the motor against the current automatic tuning action, reverse the next automatic action
            if (m_LastAction * effort < 0)
            {
                m_LastAction = -m_LastAction;
            }

			// just store it, the main loop will make it work
			m_UserPulse = new PulseNode(effort, count, true);

		}
		#endregion

		#region Basic Pulse-Based Algorithm
		/// <summary>
		/// The states of the FSM.
		/// </summary>
		private enum SimpleState {
			Initialize,
			Sampling,
			PulseStarted,
			PulseEnded,
			UserMove,
		}

		/// <summary>
		/// The last sensor reading that was considered.
		/// </summary>
		private ulong m_LastSensorScan;

		/// <summary>
		/// The measured SWR that caused the last motor movement.
		/// </summary>
		private decimal m_LastActionSWR;

		/// <summary>
		/// The last action presented to the motor.
		/// </summary>
		private decimal m_LastAction;

		/// <summary>
		/// The current motor state.
		/// </summary>
		private decimal m_CurrentAction;

		/// <summary>
		/// The frequency of the last action.
		/// </summary>
		private decimal m_LastTunedFreq;

		/// <summary>
		/// The last frequency read from the radio.
		/// </summary>
		private decimal m_LastReadFreq;

		/// <summary>
		/// The frequency-related hint for the next TUNING cycle's
		/// initial direction.
		/// </summary>
		private int FreqHint { get { return m_FreqHintCounter.BinarySign; } }

		/// <summary>
		/// When using an ITimedMotor, this is the date/time tick count for
		/// the next sensor reading.
		/// </summary>
		private long m_NextRead;

		/// <summary>
		/// Indicates that the FSM is searching for a new tune point.
		/// </summary>
		private bool m_Searching;

		/// <summary>
		/// The algorithm state.
		/// </summary>
		private SimpleState m_State;

		/// <summary>
		/// Algorithm-triggered pulse.
		/// </summary>
		private PulseNode m_AutoPulse;

		/// <summary>
		/// The user move value last time it was evaluated.
		/// </summary>
		private decimal m_LastUserMove;

		/// <summary>
		/// The initial-move threshold counter.
		/// </summary>
		private ushort m_AlarmCount;

#if VERBOSE_DEBUG
		/// <summary>
		/// For debugging.
		/// </summary>
		private SimpleState m_LastState;
#endif

		/// <summary>
		/// Worker thread for pulse/step-based algorithm.
		/// </summary>
		private void PulseAlgorithmThread() {
			// set initial FSM state
			m_State = SimpleState.Initialize;

			// use dynamic casting to detect and extract specific features
			var timed = m_Motor as ITimedMotor;
			var braked = m_Motor as IBrakedMotor;
			var stepper = m_Motor as IStepperMotor;
			var mVersion = m_Motor as IQueryVersion;
			var sVersion = m_Sensor as IQueryVersion;
			var rVersion = m_Radio as IQueryVersion;

			while (!m_Quit) {
				try {
#if VERBOSE_DEBUG
					if (m_State != m_LastState)
						Debug.WriteLine(String.Format("DEBUG: Core State = {0}", m_State));
					m_LastState = m_State;
#endif
					//
					//  STATE MACHINE
					//
					switch (m_State) {
						#region Initialization
						case SimpleState.Initialize: {
							m_LastActionSWR = 100m;
							m_LastAction = 1m;
							m_State = SimpleState.Sampling;
							m_Searching = false;
							//m_FreqHintCounter.Update(1); // start with "up" hint
							m_LastTunedFreq = 0m; // "empty"
							m_AlarmCount = 0;

							// update the SWR-only flag
							if (m_Sensor.ReadOnlySWR != Config.ReadOnlySWR) {
								m_Sensor.ReadOnlySWR = Config.ReadOnlySWR;
							}

							// start I/O objects
							foreach (var start in new[] { m_Motor as IStart, m_Sensor as IStart, m_Radio as IStart }) {
								if (!ReferenceEquals(start, null)) {
									try {
										start.Start();
									} catch (Exception ex) {
										Debug.WriteLine(String.Format("DEBUG: Initialize({0}) threw {1}: {2}", start.GetType().Name, ex.GetType().Name, ex.Message));
									}
								}
							}

							// update motor polarity flag
							m_Motor.MotorReversed = Config.MotorPolarityReversed;

							// update the braking state
							if (!ReferenceEquals(braked, null)) {
								braked.SetBraking(Config.MotorBrakingEnabled);
							}

							// set a safe state
							m_Motor.Move(0m);

						} break;
						#endregion

						#region Sampling
						case SimpleState.Sampling: {
							try {
								// update motor-reverse setting if it has changed
								if (m_Motor.MotorReversed != Config.MotorPolarityReversed) {
									m_Motor.MotorReversed = Config.MotorPolarityReversed;
									// flipping the motor polarity invalidates the hint counter
									m_FreqHintCounter.Reset();
								}

								// update the SWR-only flag
								if (m_Sensor.ReadOnlySWR != Config.ReadOnlySWR) {
									m_Sensor.ReadOnlySWR = Config.ReadOnlySWR;
								}

								// update the braking settings
								if (!ReferenceEquals(braked, null) && Config.MotorBrakingEnabled != braked.Brake) {
									braked.SetBraking(Config.MotorBrakingEnabled);
								}

								// update the version strings as needed
								if (!ReferenceEquals(mVersion, null) && ReferenceEquals(mVersion.VersionString, null)) {
									mVersion.QueryVersion();
								}
								if (!ReferenceEquals(sVersion, null) && ReferenceEquals(sVersion.VersionString, null)) {
									sVersion.QueryVersion();
								}
								if (!ReferenceEquals(rVersion, null) && ReferenceEquals(rVersion.VersionString, null)) {
									rVersion.QueryVersion();
								}

								// process a user pulse if one exists
								if (m_UserPulse != null) {
									m_AutoPulse = new PulseNode(m_UserPulse);
									m_CurrentAction = m_UserPulse.Effort;
									m_State = SimpleState.PulseStarted;
									// don't use tables for next action
									m_LastTunedFreq = 0m;
									break;
								}

								// process a user move if one exists
								var userMove = m_UserMoveEffort;
								if (userMove != null) {
									m_LastUserMove = userMove.Value;
									m_State = SimpleState.UserMove;
									m_Motor.Move(userMove.Value);
									m_CurrentAction = m_LastUserMove;
									// don't use tables for next action
									m_LastTunedFreq = 0m;
									break;
								}

								// read the frequency
								var freq = m_Radio.Frequency;
								m_LastReadFreq = freq;

								// read sensor
								if (m_Data == null) {
									break; // no data yet
								}
								var current = m_SensorVersion;
								if (m_LastSensorScan == current) {
									break; // stale data
								}
								m_LastSensorScan = current;
								if ((m_Data.Power != null) && (m_Data.Power < Config.MinPower)) {
									m_AlarmCount = 0; // reset the initial threshold counter
									break; // not enough power to tune
								}

								// if automatic mode disabled, give up now
								if (!Automatic) {
									break;
								}

								// measure SWR and hold reading
								var swr = m_Data.SWR;

								// compare against configured thresholds
								var alarm = false;
								var storeSWR = true;
								var bump = false; // alarm counter control var
								for (int i = 0; i != Config.Thresholds.Length; ++i) {
									if (swr > Config.Thresholds[i].SWR) {
										bump = true;
										if (m_AlarmCount < Config.AlarmCount)
											++m_AlarmCount;
										if (m_Searching || (m_AlarmCount >= Config.AlarmCount))
											alarm = true;
									}
#if VERBOSE_DEBUG
								if (bump || alarm || (m_AlarmCount != 0))
									Debug.WriteLine("DEBUG: bump = {0}; alarm = {1}; count = {2}", bump, alarm, m_AlarmCount);
#endif
									// only process excursion if enough 'bumps' have occurred to
									//    trigger an alarm
									if (bump && !alarm) break;

									// process alarm condition
									if (alarm) {
										// if we are already searching, use SWR for direction hints
										if (m_Searching) {
											//
											//  if we are already searching, use SWR for direction hints
											//
											var cmp = CompareSWR(swr, m_LastActionSWR, Config.SensorEpsilon);
											// if the SWR is getting worse, go the other way
											if (cmp > 0) {
												// assume last direction was wrong, and invert
												m_LastAction = -m_LastAction;
											} else if (cmp == 0) {
												// if the SWR hasn't changed enough, just keep going
												storeSWR = false;
											}
										} else if ((freq != 0m) && (m_LastTunedFreq != 0m) && (freq != m_LastTunedFreq)) {
											//
											// if this is the first SWR excursion, and frequency
											//    support is available, use that as the initial hint
											//
											if (freq < m_LastTunedFreq) {
												m_LastAction = -m_FreqHintCounter.BinarySign;
											} else {
												//
												//  if this is a small excursion, or the table is not available,
												//     use the frequency hint as the initial movement
												//
												if (freq < m_LastTunedFreq) {
													m_LastAction = -m_FreqHintCounter.BinarySign;
												} else {
													m_LastAction = m_FreqHintCounter.BinarySign;
												}

												Debug.WriteLine(String.Format("DEBUG: Get hint, freq = {0}, last = {1}, hint = {2}", freq, m_LastTunedFreq, m_LastAction));
											}

											Debug.WriteLine(String.Format("DEBUG: Get hint, freq = {0}, last = {1}, hint = {2}", freq, m_LastTunedFreq, m_LastAction));
										}

										// pulse the motor according to the configuration
										m_AutoPulse = new PulseNode(
											m_LastAction * Config.Thresholds[i].Effort,
											Config.Thresholds[i].Size,
											Config.Thresholds[i].IsStep && !ReferenceEquals(stepper, null)
										);
										m_State = SimpleState.PulseStarted;
										m_CurrentAction = m_LastAction * Config.Thresholds[i].Effort;
										m_Searching = true;
										break;
									}
								}

#if VERBOSE_DEBUG				// DEBUG
								if (alarm) {
									Debug.WriteLine("DEBUG: SWR={0}; LastSWR={1}; Action={2}; Searching={3}",
										swr, m_LastActionSWR, m_LastAction, m_Searching);
								}
#endif
								// if the SWR is satisfactory, or has changed more than Epsilon,
								//    save the SWR for the next SAMPLE cycle
								if (storeSWR) {
									m_LastActionSWR = swr;
								}

								// if SWR was OK...
								if (!bump) {
									m_AlarmCount = 0; // reset the initial threshold counter
								}

								// and maybe save the frequency for the next TUNE cycle
								if (!alarm) {
									if (m_Searching) {
										m_Searching = false; // fall out of the auto-tune cycle
										m_UserPulse = null;  // cancel any user pulses clicked while auto-tuning

										// if the VFO moved, update the frequency hint
										if ((freq != 0m) && Config.RadioLearningEnabled) {
											if (freq > m_LastTunedFreq) {
												m_FreqHintCounter.Update(m_LastAction >= 0m ? 1 : -1);
											} else if (freq < m_LastTunedFreq) {
												m_FreqHintCounter.Update(m_LastAction >= 0m ? -1 : 1);
											}
										}
										m_LastTunedFreq = freq;
#if DEBUG
										Debug.WriteLine(String.Format("DEBUG: Tune cycle complete, freq = {0}; hint = {1}", freq, m_FreqHintCounter.Value));
#endif
									}
									//if (m_LastTunedFreq == 0m)
										m_LastTunedFreq = freq;
									m_CurrentAction = 0m;
								}
							} finally {
								// if no tuning actions were taken, sleep briefly before next sample
								if (m_State == SimpleState.Sampling) {
									Thread.Sleep(IdlePollDelay);
								}
							}
						} break;
						#endregion

						#region Pulse Started
						case SimpleState.PulseStarted: {
#if DEBUG
							Debug.WriteLine("DEBUG: {0}: {1}% {2}{3}",
								(m_AutoPulse.IsStep ? "STEP" : "PULSE"),
								(int)(m_AutoPulse.Effort * 100m),
								m_AutoPulse.Size,
								(m_AutoPulse.IsStep ? "steps" : "sec")
							);
#endif
							// the motor is moving; invalidate the last tuned freq
							m_LastTunedFreq = 0m;

							// compute the expiration time
							var now = DateTimeOffset.UtcNow.UtcTicks;
							if (m_AutoPulse.IsStep) {
								// READ DELAY for STEP COUNT
								m_NextRead = now
									+ (long)(Config.ReadDelay * TimeSpan.TicksPerSecond)
									+ (long)((m_AutoPulse.Size / (Math.Abs(m_AutoPulse.Effort) * Config.MaxStepRate)) * TimeSpan.TicksPerSecond);
							} else {
								// READ DELAY for all other motors
								m_NextRead = now
									+ (long)(Config.ReadDelay * TimeSpan.TicksPerSecond)
									+ (long)(m_AutoPulse.Size * TimeSpan.TicksPerSecond);
							}
#if VERBOSE_DEBUG
							// DEBUG:
							Debug.WriteLine("DEBUG: NextRead in {0} msec", (m_NextRead - now) / (TimeSpan.TicksPerSecond / 1000));
#endif
							// choose action based on capabilities of motor...
							if (m_AutoPulse.IsStep && !ReferenceEquals(stepper, null)) {
								// command the controller to step
								stepper.Step(m_AutoPulse.Effort, Convert.ToUInt64(Math.Abs(m_AutoPulse.Size)));

								// move to next state in FSM
								m_State = SimpleState.PulseEnded;
							} else if (!ReferenceEquals(timed, null)) {
								// command the controller to pulse
								timed.Pulse(m_AutoPulse.Effort, m_AutoPulse.Size);

								// move to next state in FSM
								m_State = SimpleState.PulseEnded;
							} else { // if we provide the timing for the pulse, use a delay to generate it...
								// start the motor
								m_Motor.Move(m_AutoPulse.Effort);

								// wait for pulse duration
								if (!m_AutoPulse.IsStep)
									Thread.Sleep(Convert.ToInt32(m_AutoPulse.Size * 1000m));
								// stop motor
								m_Motor.Move(0m);
								m_CurrentAction = 0m;

								// move to next state in FSM
								m_State = SimpleState.PulseEnded;
							}
						} break;
						#endregion

						#region Pulse Ended
						case SimpleState.PulseEnded: {
							if ((!ReferenceEquals(timed, null)) || (!ReferenceEquals(stepper, null))) {
								// if time elapsed, clean up and allow the next reading
								if (DateTimeOffset.UtcNow.UtcTicks >= m_NextRead) {
									// move to next state in FSM
									m_State = SimpleState.Sampling;
									// clear all pending pulse details
									m_UserPulse = null;
									m_AutoPulse = null;
									m_CurrentAction = 0m;
								} else {
									Thread.Sleep(TimedSpinDelay);
								}
							} else {
								// wait for motor to settle
								Thread.Sleep(Convert.ToInt32(Config.ReadDelay * 1000.0m));
								// move to next state in FSM
								m_State = SimpleState.Sampling;
								// clear all pending pulse details
								m_UserPulse = null;
								m_AutoPulse = null;
								m_CurrentAction = 0m;
							}
						} break;
						#endregion

						#region User Move
						case SimpleState.UserMove: {
							var userMove = m_UserMoveEffort;
							// if the user has canceled, then stop moving
							if (userMove == null || userMove.Value == 0m) {
								m_UserMoveEffort = null;
								m_LastUserMove = 0m;
								m_CurrentAction = m_LastUserMove;
								m_Motor.Move(0m);
								m_State = SimpleState.Sampling;
								break;
							}
							// otherwise, reevaluate the effort
							if (userMove.Value != m_LastUserMove) {
								m_LastUserMove = userMove.Value;
								m_Motor.Move(userMove.Value);
							}

							// store what we're doing
							m_CurrentAction = m_LastUserMove;

							// take a breath
							Thread.Sleep(TimedSpinDelay);
						} break;
						#endregion
					}
				} catch (Exception ex) {
					Debug.WriteLine(String.Format("DEBUG: {0}: {1}: {2}",
						ex.GetType().Name,
						ex.Message,
						ex.StackTrace));
				}
			}

			// shut down the sensor
			m_Sensor.AutoPoll = false;

			// make sure the motor is STOPPED
			m_Motor.Move(0m);
		}

		/// <summary>
		/// Compare SWR against last action, to see if the drift is enough to continue acting.
		/// </summary>
		/// <param name="swr">The current SWR measurement.</param>
		/// <param name="m_LastActionSWR">The SWR from the last action.</param>
		/// <returns>-1 -> current SWR is lower, 0 -> current SWR is roughly equal, 1 -> current SWR is higher.</returns>
		private static int CompareSWR(decimal current, decimal previous, decimal epsilon) {
			var difference = current - previous;
			if (difference > epsilon) {
#if VERBOSE_DEBUG
				Debug.WriteLine(String.Format("DEBUG: Compare SWR: current={0:0.00}, previous={1:0.00}, epsilon={2:0.00}: GREATER", current, previous, epsilon));
#endif
				return 1;
			}
			if (difference < -epsilon) {
#if VERBOSE_DEBUG
				Debug.WriteLine(String.Format("DEBUG: Compare SWR: current={0:0.00}, previous={1:0.00}, epsilon={2:0.00}: LESS", current, previous, epsilon));
#endif
				return -1;
			}
#if VERBOSE_DEBUG
			Debug.WriteLine(String.Format("DEBUG: Compare SWR: current={0:0.00}, previous={1:0.00}, epsilon={2:0.00}: EQUAL", current, previous, epsilon));
#endif
			return 0;
		}
		#endregion

		#region Event Handlers & Utilities
#if false
		/// <summary>
		/// Worker method for pulse async call.
		/// </summary>
		/// <param name="node">The pulse details.</param>
		private void PulseWorker(PulseNode node) {
			m_Motor.Move(node.Effort);
			Thread.Sleep(Convert.ToInt32(node.Duration * 1000m));
			m_Motor.Move(0);
			m_UserPulse = null;
		}

		/// <summary>
		/// Callback for pulse completion.
		/// </summary>
		/// <param name="ar">The system callback block.</param>
		private void PulseCallback(IAsyncResult ar) {
			// nop for now
		}
#endif

		/// <summary>
		/// Raised when new sensor data is available.
		/// </summary>
		private void SensorDataReceived(object sender, SensorDataEvent e) {
			m_Data = e.Data;
			++m_SensorVersion;
#if DEBUG
			if (m_SensorVersion % 100 == 0) {
				var now = DateTimeOffset.UtcNow;
				var rate = 100.0 / (now - m_LastSensorRateUpdate).TotalSeconds;
				Debug.WriteLine("DEBUG: Sensor Data Version {0} ({1:0.0} Hz)", m_SensorVersion, rate);
				m_LastSensorRateUpdate = now;
			}
#endif
			//Debug.WriteLine("DEBUG: New SWR = {0:0.00}, Power = {1:0.0} ({2})", e.Data.SWR, e.Data.Power, m_SensorVersion);
		}
		#endregion
	}
}
