﻿/*
 *
 *
 *    GenericBase.cs
 *
 *    AutoCap - Base type for generic message endpoints.
 *
 *    License: GNU General Public License Version 3.0.
 *    
 *    Copyright (C) 2014 by Matthew K. Roberts, KK5JY. All rights reserved.
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *    
 *    This program is distributed in the hope that it will be useful, but
 *    WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    General Public License for more details.
 *    
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see: http://www.gnu.org/licenses/
 *    
 *
 */

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Net.Sockets;
using System.Threading;
using System.Net;

namespace KK5JY.AutoCap {
	/// <summary>
	/// A parsed message.
	/// </summary>
	public sealed class GenericMessage : Dictionary<string, string> {
		public GenericMessage()
			: base(StringComparer.InvariantCultureIgnoreCase) {
			// nop
		}
	}

	/// <summary>
	/// The message event.
	/// </summary>
	public sealed class GenericMessageEvent : EventArgs {
		/// <summary>
		/// The message.
		/// </summary>
		public GenericMessage Message { get; internal set; }
	}

	/// <summary>
	/// An event indicating a malformed message was sent.
	/// </summary>
	public sealed class GenericMessageError : EventArgs {
		/// <summary>
		/// The message text, if available.
		/// </summary>
		public string Data { get; internal set; }
	}

	/// <summary>
	/// A class for generic socket interfaces to everything.
	/// </summary>
	/// <remarks>
	/// The protocol used is a CSV set of key=value pairs.  These can appear
	/// in any quantity on a single line, terminated by CRLF.  Each single line
	/// is a single command, query, or response.  Both peers use this scheme to
	/// trade data and commands over a single socket link.  Classes that use
	/// this type for communication should document the specific KVPs needed
	/// to properly function.
	/// </remarks>
	public class GenericBase : IDisposable {
		public const string CRLF = "\r\n";

		#region Private Fields
		private Socket m_Sock;
		private byte[] m_IoBuffer;
		private string m_RxBuffer;
		#endregion

		#region Public Properties and Events
		/// <summary>
		/// True if the socket has closed.
		/// </summary>
		public bool IsClosed { get; private set; }

		/// <summary>
		/// Raised when a properly formatted message is received.
		/// </summary>
		public event EventHandler<GenericMessageEvent> MessageReceived;

		/// <summary>
		/// Raised when a malformed message is received.
		/// </summary>
		public event EventHandler<GenericMessageError> MalformedMessage;
		#endregion

		#region Public Methods
		/// <summary>
		/// Create a new socket handler.
		/// </summary>
		/// <param name="s">The socket.</param>
		public GenericBase(Socket s) {
			m_Sock = s;
			IsClosed = true;
			m_RxBuffer = "";
			m_IoBuffer = new byte[1024];
			s.UseOnlyOverlappedIO = true;
		}

		/// <summary>
		/// Start I/O.
		/// </summary>
		public void Open() {
			if (IsClosed) {
				StartRx();
			}
		}

		/// <summary>
		/// Close the socket.
		/// </summary>
		public void Dispose() {
			Close();
		}

		/// <summary>
		/// Close the socket.
		/// </summary>
		public void Close() {
			IsClosed = true;
			try { m_Sock.Close(); } catch { /* nop */ }
		}

		/// <summary>
		/// Send a message to the peer.
		/// </summary>
		/// <param name="msg">The message to send.</param>
		public void SendMessage(GenericMessage msg) {
			string toSend = "";
			foreach (var kvp in msg) {
				if (toSend.Length != 0) {
					toSend += ",";
				}
				toSend += String.Format("{0}={1}", kvp.Key, kvp.Value);
			}
			toSend += CRLF;
			for (int i = 0; i != m_IoBuffer.Length && i != toSend.Length; ++i) {
				m_IoBuffer[i] = (byte)(toSend[i]);
			}
			m_Sock.Send(m_IoBuffer, 0, toSend.Length, SocketFlags.None);
		}
		#endregion

		#region Event Handlers
		/// <summary>
		/// Raised when a properly formatted message is received.
		/// </summary>
		protected virtual void OnMessageReceived(GenericMessageEvent e) {
			if (MessageReceived != null) {
				try {
					MessageReceived(this, e);
				} catch {
					// nop
				}
			}
		}

		/// <summary>
		/// Raised when a malformed message is received.
		/// </summary>
		protected virtual void OnMalformedMessage(GenericMessageError e) {
			if (MalformedMessage != null) {
				try {
					MalformedMessage(this, e);
				} catch {
					// nop
				}
			}
		}
		#endregion

		#region Utilities
		private void StartRx() {
			try {
				m_Sock.BeginReceive(m_IoBuffer, 0, m_IoBuffer.Length, SocketFlags.None, RxCallback, null);
			} catch {
				IsClosed = true;
			}
		}

		private void RxCallback(IAsyncResult ar) {
			try {
				var len = m_Sock.EndReceive(ar);
				if (len <= 0) {
					IsClosed = true;
				}
				for (int i = 0; i != len; ++i) {
					m_RxBuffer += (char)(m_IoBuffer[i]);
				}
				ProcessRxBuffer();
			} catch {
				IsClosed = true;
			} finally {
				if (!IsClosed)
					StartRx();
			}
		}

		private void ProcessRxBuffer() {
			int loc = m_RxBuffer.IndexOf(CRLF);
			while (loc >= 0) {
				// extract the first message
				var data = m_RxBuffer.Substring(0, loc + 2);
				m_RxBuffer = m_RxBuffer.Substring(loc + 2);

				// parse the extracted message text
				try {
					var msg = new GenericMessage();
					var pairs = data.Split(new[] { ',' }, StringSplitOptions.RemoveEmptyEntries);
					foreach (var pair in pairs) {
						var tokens = pair.Split(new[] { '=' }, 2);
						if (!String.IsNullOrEmpty(tokens[0]) && !String.IsNullOrEmpty(tokens[1])) {
							msg.Add(tokens[0], tokens[1]);
						}
						OnMessageReceived(new GenericMessageEvent { Message = msg });
					}
				} catch {
					OnMalformedMessage(new GenericMessageError { Data = data });
				}

				// advance to next message
				loc = m_RxBuffer.IndexOf(CRLF);
			}
		}
		#endregion
	}

	/// <summary>
	/// Actively connects to a remote host.
	/// </summary>
	public abstract class GenericConnector {
		#region Private Fields
		private GenericBase m_Worker;
		private Timer m_Timer;
		private bool m_TimerWorking;
		#endregion

		#region Public Properties
		/// <summary>
		/// The endpoint address of the sensor.
		/// </summary>
		public IPEndPoint EndPoint { get; private set; }
		#endregion

		#region Public Methods
		/// <summary>
		/// Create a new connector.
		/// </summary>
		/// <param name="ep">The remote endpoint address.</param>
		public GenericConnector(IPEndPoint ep) {
			EndPoint = ep;
		}

		/// <summary>
		/// Start I/O.
		/// </summary>
		public void Open() {
			m_Timer = new Timer(TimerTick, null, 0, 5000);
		}

		/// <summary>
		/// Close the sensor.
		/// </summary>
		public void Stop() {
			Dispose();
		}

		/// <summary>
		/// Close the sensor.
		/// </summary>
		public void Dispose() {
			m_Timer.Dispose();
			var worker = m_Worker;
			if (worker != null) {
				worker.Dispose();
			}
			m_Worker = null;
		}

		/// <summary>
		/// Send a message to the remote peer.
		/// </summary>
		/// <param name="msg">The message.</param>
		public bool SendMessage(GenericMessage msg) {
			var worker = m_Worker;
			if (worker != null) {
				try {
					worker.SendMessage(msg);
					return true;
				} catch {
					// nop
				}
			}
			return false;
		}
		#endregion

		#region Event Handlers
		private void TimerTick(object state) {
			if (m_TimerWorking) return;
			m_TimerWorking = true;
			try {
				var w = m_Worker;
				if (w.IsClosed) {
					w.MessageReceived -= MessageReceived;
					w.MalformedMessage -= MalformedMessage;
					w = m_Worker = null;
				}
				if (w == null) {
					var sock = new Socket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp);
					sock.Connect(EndPoint);
					m_Worker = new GenericBase(sock);
					m_Worker.MessageReceived += MessageReceived;
					m_Worker.MalformedMessage += MalformedMessage;
					m_Worker.Open();
				}
			} catch {
				// nop
			}
			m_TimerWorking = false;
		}

		private void MalformedMessage(object sender, GenericMessageError e) {
			OnMalformedMessage(e);
		}

		private void MessageReceived(object sender, GenericMessageEvent e) {
			OnMessageReceived(e);
		}

		protected abstract void OnMalformedMessage(GenericMessageError e);

		protected abstract void OnMessageReceived(GenericMessageEvent e);
		#endregion
	}
}
