﻿/*
 *
 *
 *    Interfaces.cs
 *
 *    AutoCap - Some interface abstractions.
 *
 *    License: GNU General Public License Version 3.0.
 *    
 *    Copyright (C) 2014 by Matthew K. Roberts, KK5JY. All rights reserved.
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *    
 *    This program is distributed in the hope that it will be useful, but
 *    WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    General Public License for more details.
 *    
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see: http://www.gnu.org/licenses/
 *    
 *
 */

using System;

namespace KK5JY.IO {
	/// <summary>
	/// Any object that needs to be started before it will perform
	/// its task.
	/// </summary>
	public interface IStart {
		void Start();
	}
}

namespace KK5JY.AutoCap {
	/// <summary>
	/// A SWR reading event.
	/// </summary>
	public class SensorDataEvent : EventArgs {
		/// <summary>
		/// The latest reading.
		/// </summary>
		public SensorData Data { get; set; }
	}

	/// <summary>
	/// A single sensor reading.
	/// </summary>
	public class SensorData {
		/// <summary>
		/// The sensor ID, if available.
		/// </summary>
		public string ID { get; set; }

		/// <summary>
		/// The SWR reading.
		/// </summary>
		public decimal SWR { get; set; }

		/// <summary>
		/// The measured power output.
		/// </summary>
		public decimal ? Power { get; set; }

		/// <summary>
		/// The measured magnitude of the impedence, if available..
		/// </summary>
		public decimal ? Z { get; set; }

		/// <summary>
		/// The measured reactive part of the impedence, if available..
		/// </summary>
		public decimal ? X { get; set; }

		/// <summary>
		/// The measured phase angle of the impedence, if available.
		/// </summary>
		public decimal ? Phi { get; set; }
	}

	/// <summary>
	/// Reports a stepper position.
	/// </summary>
	public sealed class PositionEventArgs : EventArgs {
		/// <summary>
		/// The current stepper position.
		/// </summary>
		public long Position { get; private set; }

		/// <summary>
		/// Create a new event.
		/// </summary>
		/// <param name="pos">The current stepper position.</param>
		public PositionEventArgs(long pos) {
			Position = pos;
		}
	}

	/// <summary>
	/// Any item that reports a version string.
	/// </summary>
	public interface IReportsVersion {
		/// <summary>
		/// Returns the version string from the device.
		/// </summary>
		string VersionString { get; }
	}

	/// <summary>
	/// This is a special version of IReportsVersion that allows for
	/// the user to actively query the object for an updated version
	/// string.
	/// </summary>
	public interface IQueryVersion : IReportsVersion {
		/// <summary>
		/// Request an update to the version string.
		/// </summary>
		bool QueryVersion();
	}

	/// <summary>
	/// A generic sensor.
	/// </summary>
	public interface ISensor {
		/// <summary>
		/// Raised when the sensor receives a reading.
		/// </summary>
		event EventHandler<SensorDataEvent> SensorData;

		/// <summary>
		/// Indicates that the sensor should only try to read SWR (not power, reactance, etc.)
		/// </summary>
		bool ReadOnlySWR { get; set; }

		/// <summary>
		/// Enables automatic best-effort polling within the object.
		/// </summary>
		bool AutoPoll { get; set; }

		/// <summary>
		/// Manually poll the sensor for the next SWR value.
		/// </summary>
		void Poll();
	}

	/// <summary>
	/// A generic motor.
	/// </summary>
	public interface IMotor {
		/// <summary>
		/// Reverse the motor effort for movement commands.
		/// </summary>
		bool MotorReversed { get; set; }

		/// <summary>
		/// Move or stop the motor.
		/// </summary>
		/// <param name="speed">The normalized speed, including direction sign.</param>
		/// <returns>True if the command was successfully sent.</returns>
		bool Move(decimal speed = 0);
	}

	/// <summary>
	/// A generic stepper motor.
	/// </summary>
	public interface IStepperMotor : IMotor {
		/// <summary>
		/// An event for position reports.
		/// </summary>
		event EventHandler<PositionEventArgs> PositionUpdate;

		/// <summary>
		/// Step forward or backward (negative).
		/// </summary>
		/// <param name="speed">The normalized stepping rate (negative is reversed).</param>
		/// <param name="steps">The number of steps.</param>
		/// <returns>True if the command was successfully sent.</returns>
		bool Step(decimal speed, ulong steps);

		/// <summary>
		/// Go to a specific stepper position.
		/// </summary>
		/// <param name="target">The absolute position.</param>
		/// <returns>True iff message successfully sent.</returns>
		bool Seek(long target);

		/// <summary>
		/// Reset the stepper's 'odometer' position to zero.
		/// </summary>
		/// <returns>True iff message successfully sent.</returns>
		bool ResetPosition();

		/// <summary>
		/// Get the current stepper position.
		/// </summary>
		/// <returns>True iff message successfully sent.</returns>
		bool QueryPosition();
	}

	/// <summary>
	/// A motor that can generate precision-timed pulses.
	/// </summary>
	public interface ITimedMotor : IMotor {
		/// <summary>
		/// Generate a movement pulse.
		/// </summary>
		/// <param name="speed">The normalized effort (negative is reversed).</param>
		/// <param name="duration">The duration, in seconds.</param>
		/// <returns>True if the command was successfully sent.</returns>
		bool Pulse(decimal speed, decimal duration);
	}

	/// <summary>
	/// A braked motor.
	/// </summary>
	public interface IBrakedMotor : IMotor {
		/// <summary>
		/// The current brake state in the controller.
		/// </summary>
		bool Brake { get; }

		/// <summary>
		/// Enable (true) or disable (false) braking.
		/// </summary>
		/// <param name="enabled">The new braking feature state.</param>
		/// <returns>True if the command was successfully sent.</returns>
		bool SetBraking(bool enabled);
	}

	/// <summary>
	/// A generic radio.
	/// </summary>
	public interface IRadio {
		/// <summary>
		/// Reads the frequency from the radio.
		/// </summary>
		decimal Frequency { get; }
	}
}
