﻿/*
 *
 *
 *    MainForm.cs
 *
 *    AutoCap - Windows Form.
 *
 *    License: GNU General Public License Version 3.0.
 *    
 *    Copyright (C) 2014-2018 by Matthew K. Roberts, KK5JY.
 *    All rights reserved.
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *    
 *    This program is distributed in the hope that it will be useful, but
 *    WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    General Public License for more details.
 *    
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see: http://www.gnu.org/licenses/
 *    
 *
 */

#if DEBUG
#define VERBOSE_DEBUG
#endif

using System;
using System.Net;
using System.Linq;
using System.Drawing;
using System.Windows.Forms;
using System.Globalization;
using System.Collections.Generic;
using System.Reflection;
using System.Diagnostics;
using System.Xml;

//
//  TODO:
//      0.  Detect when ReadDelay is too small, and adjust it upwards.  A counter
//          could be used for this, to count the number of consecutive SWR rises
//          that occur, and after some number of these, assume that ReadDelay is
//          too small.
//      1.  Detect when the tuning algorithm has worked into a local minimum that
//          is still above the lowest threshold.  Another counter could be used
//          here to count how many times the SWR has cycled without actually
//          reaching the bottom threshold.
//		2.  Integrate and test at least one of { FT-8*7, TS-480, K3 }
//		3.  Add generic { socket, serial } interfaces for { motor, sensor, radio },
//			using minimal textual protocols.  This allows users to extend the
//			application without changing its code.
//

namespace KK5JY.AutoCap {
	/// <summary>
	/// Main application form.
	/// </summary>
	public partial class MainForm : Form {
		#region Constants
		private const string SerialComboBoxDefault = "None";

		private const int MessageInterval = 2; // sec

		private const int DisplayTimerInterval = 200; // msec

		private const string PulseTimeText = "ms";
		private const string PulseStepText = "st";

		/// <summary>
		/// The maximum number of ports supported by external AutoCap controllers.
		/// </summary>
		private const int MaxMotorPorts = 4;
		#endregion

		#region Inner Types
		private delegate void StringDelegate(string s);
		#endregion

		#region Private Fields
		// primary components
		private Core m_Core;
		private Timer m_Timer;

		// error message handling
		private Dictionary<string, string> m_Messages;
		private Dictionary<string, string> m_Versions;
		private int m_CurrentMessage;
		private int m_CurrentVersion;
		private int m_MessageTimer;

		// host name validation
		private string m_LastRadioHost;
		private bool m_LastRadioHostValid;
		private string m_LastRadioAddress;

		// port name pool
		private readonly List<string> m_PortNames;

		// sim controls for testing
		private SimulatorControls m_SimControls;

		// user move latching
		private bool m_LatchingMoves;

		/// <summary>
		/// Enable/disable the pulse row.
		/// </summary>
		private bool PulseRowEnabled {
			get {
				return numPulse1.Enabled;
			}
			set {
				numPulse1.Enabled = numPulse2.Enabled = numPulse3.Enabled = numPulse4.Enabled = value;
			}
		}

		/// <summary>
		/// Tracks whether the core is currently enabled (backing var).
		/// </summary>
		private bool m_CoreEnabled;

		/// <summary>
		/// Tracks whether the core is currently enabled.
		/// </summary>
		private bool CoreEnabled {
			get { return m_CoreEnabled; }
			set {
				if (value != m_CoreEnabled) {
					m_CoreEnabled = value;
					AutoTuningEnabledChanged(this, new EventArgs());
				}
				bEnable.BackColor = m_CoreEnabled ? Color.Lime : SystemColors.Control;
				bEnable.Text = m_CoreEnabled ? "Disable" : "Enable";
			}
		}
		#endregion

		#region Constructors
		/// <summary>
		/// Default ctor.
		/// </summary>
		public MainForm() {
			// initialize the designer stuff
			InitializeComponent();
			Text = Platform.VersionString;

			// initialize the core
			m_Core = new Core();

			// create the message queue for the status bar
			m_Messages = new Dictionary<string, string>();
			m_Versions = new Dictionary<string, string>();
			m_LastRadioHostValid = true;
			m_MessageTimer = 0;

			// build the list of serial ports
			m_PortNames = new List<string>(KK5JY.IO.SerialPortHelper.GetPortNames());

			// populate all of the serial port lists
			foreach (var cb in new[] { cbSensorPort, cbMotorComPort }) {
                cb.DropDownStyle = ComboBoxStyle.DropDown;
                cb.Items.Add(SerialComboBoxDefault);
				foreach (var port in m_PortNames) {
					cb.Items.Add(port);
				}
				cb.SelectedIndex = 0;
			}

			// populate the motor output port box
			for (int i = 1; i <= MaxMotorPorts; ++i) {
				cbMotorOutputPort.Items.Add(i.ToString(CultureInfo.InvariantCulture));
			}
			cbMotorOutputPort.DropDownStyle = ComboBoxStyle.DropDownList;
			cbMotorOutputPort.SelectedIndex = 0;

			// shake up the event handlers to make sure all the proper selections are made
			MotorTypeChanged(this, new EventArgs());
			SelectedSensorChanged(this, new EventArgs());
			SelectedRadioSourceChanged(this, new EventArgs());
#if DEBUG
			m_SimControls = new SimulatorControls(TestHardware.Instance);
			bSimControls.Visible = bSimControls.TabStop = rbMotorSim.Checked || rbSensorSim.Checked || rbRadioSim.Checked; ;
			rbSensorSim.Visible = true;
			rbMotorSim.Visible = true;
			rbRadioSim.Visible = true;
#endif
			// clean up designer stuff
			tsVersionValue.Text = "";
			tsStatus.Text = "";
			#if ! __MonoCS__
			numPulse1.Tag = lStepUnit1;
			numPulse1.ContextMenuStrip = this.contextMenuStrip1;
			numPulse2.Tag = lStepUnit2;
			numPulse2.ContextMenuStrip = this.contextMenuStrip1;
			numPulse3.Tag = lStepUnit3;
			numPulse3.ContextMenuStrip = this.contextMenuStrip1;
			numPulse4.Tag = lStepUnit4;
			numPulse4.ContextMenuStrip = this.contextMenuStrip1;
			#endif
		}
		#endregion

		#region Event Handlers
		/// <summary>
		/// Raised when the window is first shown.
		/// </summary>
		protected override void OnShown(EventArgs e) {
			if (m_Timer == null) {
				m_Timer = new Timer();
				m_Timer.Interval = DisplayTimerInterval;
				m_Timer.Tick += TimerTick;
				m_Timer.Enabled = true;
			}

			// select default radio buttons
			rbMotor232.Checked = true;
			rbSensorLP100A.Checked = true;
			rbRadioNone.Checked = true;
			bEnable.Select();

			// validate the serial ports
			ValidatePorts();

			// validate the timings
			ValidateTimings();
				
			// load persisted settings
			LoadConfig();

			// chain to base
			base.OnShown(e);
		}

		/// <summary>
		/// Raised when the form starts to close.
		/// </summary>
		protected override void OnClosing(System.ComponentModel.CancelEventArgs e) {
			// stop the core
			m_Core.Stop();

			// stop the update timer
			var timer = m_Timer;
			if (timer != null) {
				timer.Enabled = false;
				m_Timer = null;
			}

			// persist current settings
			SaveConfig();

			// chain to base
			base.OnClosing(e);
		}

		/// <summary>
		/// Raised periodically to update different parts of the UI.
		/// </summary>
		private void TimerTick(object sender, EventArgs e) {
			//
			//  Host Name Check - if radio enabled, make sure the host name is valid
			//
			if ((m_Core.Config.RadioType != RadioTypes.None) && (m_Core.Config.RadioType != RadioTypes.Sim) &&
				!String.Equals(cbRadioHost.Text, m_LastRadioHost, StringComparison.InvariantCultureIgnoreCase)) {
				m_LastRadioHost = cbRadioHost.Text;
				StringDelegate d = new StringDelegate(ValidateRadioSourceAddress);
				try {
					d.BeginInvoke(cbRadioHost.Text, null, null);
				} catch {
					// nop
				}
			}
			if (m_LastRadioHostValid || m_Core.Config.RadioType == RadioTypes.None) {
				m_Messages["RadioHost"] = ""; // no error message
				if (m_Core.Config.RadioHost != m_LastRadioAddress &&
					m_Core.Config.RadioType != RadioTypes.None)
					m_Core.Config.RadioHost = m_LastRadioAddress;
			} else {
				m_Messages["RadioHost"] = "Radio Connection Host Name Invalid";
			}

			//
			//  Rotate Messages
			//
			if (m_MessageTimer-- == 0) {
				// rotate error strings
				m_MessageTimer = MessageInterval * (1000 / DisplayTimerInterval);
				var entry = m_CurrentMessage;
				do {
					var msg = m_Messages.ElementAt(m_CurrentMessage).Value;
					m_CurrentMessage = (m_CurrentMessage + 1) % m_Messages.Count;
					if (!String.IsNullOrEmpty(msg)) {
						tsStatus.Text = msg;
						tsStatus.BackColor = Color.Salmon;
						return;
					}
				} while (m_CurrentMessage != entry);
				tsStatus.Text = "";
				tsStatus.BackColor = SystemColors.Control;
				m_CurrentMessage = 0;

				// rotate version strings
				m_Versions["Motor"] = m_Core.MotorVersion;
				m_Versions["Sensor"] = m_Core.SensorVersion;
				m_Versions["Radio"] = m_Core.RadioVersion;

				bool found = false;
				for (int i = 0; i != 3; ++i) {
					m_CurrentVersion++;
					m_CurrentVersion %= 3;
					var key = m_Versions.Keys.ElementAt(m_CurrentVersion);
					var val = m_Versions[key];
					if (!String.IsNullOrEmpty(val)) {
						found = true;
						tsVersionLabel.Text = key + ":";
						tsVersionValue.Text = val;
						break;
					}
				}
				tsVersionLabel.Visible = tsVersionValue.Visible = found;
			}

			//
			//  Update SWR, Frequency, Motor status
			//
			if (CoreEnabled) {
				if (m_Core.Data != null) {
					lSWR.Text = String.Format(CultureInfo.InvariantCulture, "{0:F2}", m_Core.Data.SWR);
					if (m_Core.Data.Power != null) {
						lPower.Text = String.Format(CultureInfo.InvariantCulture, "{0:F1} W", m_Core.Data.Power); // Power in W
					} else {
						lPower.Text = "n/a";
					}
				}
				lFreq.Text = FormatFrequency(m_Core.Frequency); // freq in MHz
				var motor = "";
				if (m_Core.MotorEffort > 0) {
					motor = String.Format(CultureInfo.InvariantCulture, "{0:0}% UP", m_Core.MotorEffort * 100m);
					bUP.BackColor = Color.Lime;
					bDown.BackColor = SystemColors.Control;
				} else if (m_Core.MotorEffort < 0) {
					motor = String.Format(CultureInfo.InvariantCulture, "{0:0}% DN", m_Core.MotorEffort * -100m);
					bDown.BackColor = Color.Lime;
					bUP.BackColor = SystemColors.Control;
				} else {
					motor = "STOPPED";
					bDown.BackColor = SystemColors.Control;
					bUP.BackColor = SystemColors.Control;
				}
				lMotor.Text = motor;
			} else {
				lSWR.Text = "n/a";
				lFreq.Text = "n/a";
				bUP.BackColor = SystemColors.Control;
				bDown.BackColor = SystemColors.Control;
			}
		}

		/// <summary>
		/// Raised when the 'sim controls' button is clicked.
		/// </summary>
		private void SimControlsClicked(object sender, EventArgs e) {
			// toggle the sim controls window visibility
			if (m_SimControls.Visible) {
				m_SimControls.Hide();
			} else {
				m_SimControls.Show();
				m_SimControls.Location = new Point(this.Location.X + this.Width + 5, this.Location.Y);
			}
		}

		/// <summary>
		/// Raised when timings are changed.
		/// </summary>
		private void ConfigurationTimingsChanged(object sender, EventArgs e) {
			ValidateTimings();
		}

		/// <summary>
		/// Raised when the Enable button is clicked.
		/// </summary>
		private void EnableClick(object sender, EventArgs e) {
			CoreEnabled = !CoreEnabled;
		}

		/// <summary>
		/// Raised when the Pause button is clicked.
		/// </summary>
		private void PauseClick(object sender, EventArgs e) {
			m_Core.Automatic = !m_Core.Automatic;
			bPause.BackColor = !m_Core.Automatic ? Color.Yellow : SystemColors.Control;
		}

		/// <summary>
		/// Raised when the enabled checkbox changed.
		/// </summary>
		private void AutoTuningEnabledChanged(object sender, EventArgs e) {
			// clear the readings
			lSWR.Text = "n/a";
			lPower.Text = "n/a";
			lFreq.Text = "n/a";
			lMotor.Text = "n/a";

			// adjust control availability
			bUP.Enabled = bDown.Enabled
				= bFastUp.Enabled = bFastDown.Enabled
				= lManualPulse.Enabled = numManualPulse.Enabled
				= lManualEffort.Enabled = numManualEffort.Enabled
				= CoreEnabled;
			bLoadProfile.Enabled = bCreateProfile.Enabled
				= rbMotor232.Enabled = rbMotor232Bridge.Enabled
				= rbMotorAutoCapV1.Enabled = rbMotorSim.Enabled
				= rbRadioFLDigi.Enabled = rbRadioHRD.Enabled
				= rbRadioNone.Enabled = rbRadioSim.Enabled
				= rbSensorAutoCapV1.Enabled = rbSensorLP100A.Enabled = rbSensorSim.Enabled
				= !CoreEnabled;

			if (CoreEnabled) {
				try {
					m_Core.Start();
#if DEBUG
				} catch (Exception ex) {
#else
				} catch {
#endif
					var msg = "Could not enable automatic tuning.  Please check your settings.";
#if DEBUG
					msg += "\r\n";
					msg += String.Format("{0}: {1}: {2}", ex.GetType().Name, ex.Message, ex.StackTrace);
#endif
					MessageBox.Show(msg, "Error");
					CoreEnabled = false;
				}
			} else {
				m_Core.Stop();
			}
		}

		/// <summary>
		/// Raised when the sensor type changes.
		/// </summary>
		private void SelectedSensorChanged(object sender, EventArgs e) {
			if (rbSensorAutoCapV1.Checked) m_Core.Config.SensorType = SensorTypes.AutoCapV1;
			if (rbSensorLP100A.Checked) m_Core.Config.SensorType = SensorTypes.LP100A;
			if (rbSensorSim.Checked) m_Core.Config.SensorType = SensorTypes.Sim;
			ValidatePorts();
#if DEBUG
			Debug.WriteLine(String.Format("DEBUG: Set Sensor Type = {0}", m_Core.Config.SensorType));
			cbSensorPort.Enabled = !rbSensorSim.Checked;
			bSimControls.Visible = bSimControls.TabStop = rbMotorSim.Checked || rbSensorSim.Checked || rbRadioSim.Checked;
#else
			cbSensorPort.Enabled = true;
#endif
		}

		/// <summary>
		/// Raised when the motor type changes.
		/// </summary>
		private void MotorTypeChanged(object sender, EventArgs e) {
			if (rbMotor232.Checked) m_Core.Config.MotorType = MotorTypes.RS232;
            if (rbMotor232Bridge.Checked) m_Core.Config.MotorType = MotorTypes.RS232Bridge;
			if (rbMotorAutoCapV1.Checked) m_Core.Config.MotorType = MotorTypes.AutoCapV1;
			if (rbMotorSim.Checked) m_Core.Config.MotorType = MotorTypes.Sim;
            Debug.WriteLine(String.Format("DEBUG: Set Motor Type = {0}", m_Core.Config.MotorType));

			ValidatePorts();

			bSimControls.Visible = bSimControls.TabStop = rbMotorSim.Checked || rbSensorSim.Checked || rbRadioSim.Checked; ;
			cbMotorComPort.Enabled = !rbMotorSim.Checked;
			cbMotorOutputPort.Enabled = rbMotorAutoCapV1.Checked;
			cbMotorRelayReversed.Enabled = rbMotor232.Checked;
			cbMotorBraking.Enabled = rbMotorAutoCapV1.Checked;
            if (m_Core.Config.MotorType != MotorTypes.AutoCapV1) {
                foreach (var label in new[] { lStepUnit1, lStepUnit2, lStepUnit3, lStepUnit4 }) {
                    SetStepChecked(label, false);
                }
                ValidateTimings();
            }
            ManualStepCheckChanged(lManualPulseMode, new EventArgs());
        }

		/// <summary>
		/// Raised when the radio source text is changed.
		/// </summary>
		private void SelectedRadioSourceChanged(object sender, EventArgs e) {
			RadioTypes newType = RadioTypes.None;
			if (rbRadioNone.Checked) {
				newType = RadioTypes.None;
				lRadioHost.Enabled =
				cbRadioHost.Enabled = false;
			} else if (rbRadioHRD.Checked) {
				newType = RadioTypes.HRD;
				lRadioHost.Enabled =
				cbRadioHost.Enabled = true;
			} else if (rbRadioFLDigi.Checked) {
				newType = RadioTypes.FLDigi;
				lRadioHost.Enabled =
				cbRadioHost.Enabled = true;
			} else if (rbRadioSim.Checked) {
				newType = RadioTypes.Sim;
				lRadioHost.Enabled =
				cbRadioHost.Enabled = false;
			}
			if (m_Core.Config.RadioType != newType) {
				m_Core.Config.RadioType = newType;
			}
			Debug.WriteLine(String.Format("DEBUG: Set Radio Type = {0}", m_Core.Config.RadioType));
		}

		/// <summary>
		/// Raised when any of the 'step' labels is clicked.
		/// </summary>
		private void StepCheckChanged(object sender, EventArgs e) {
			switch (m_Core.Config.MotorType) {
				case MotorTypes.AutoCapV1: break;
				case MotorTypes.Sim: break;
				default: return;
			}
			var label = sender as Label;
			if (!ReferenceEquals(label, null)) {
				label.Text = label.Text.Equals(PulseTimeText, StringComparison.InvariantCultureIgnoreCase) ? PulseStepText : PulseTimeText;
			}
			ValidateTimings();
		}

		/// <summary>
		/// Raised when the manual timing label is clicked.
		/// </summary>
		private void ManualStepCheckChanged(object sender, EventArgs e) {
			switch (m_Core.Config.MotorType) {
                case MotorTypes.AutoCapV1:
                case MotorTypes.Sim: {
                    var label = sender as Label;
                    if (!ReferenceEquals(label, null))
                    {
                        label.Text = label.Text.Equals(PulseTimeText, StringComparison.InvariantCultureIgnoreCase) ? PulseStepText : PulseTimeText;
                    }
                } break;
                default: {
                    var label = sender as Label;
                    if (!ReferenceEquals(label, null))
                    {
                        label.Text = PulseTimeText;
                    }
                    break;
                }
            }
		}

		/// <summary>
		/// Raised when the LP-100 port changes.
		/// </summary>
		private void SensorPortChanged(object sender, EventArgs e) {
			ValidatePorts();
		}

		/// <summary>
		/// Raised when the motor serial port changes.
		/// </summary>
		private void MotorComPortChanged(object sender, EventArgs e) {
			ValidatePorts();
		}

		/// <summary>
		/// Raised when the read delay is changed.
		/// </summary>
		private void ReadDelayChanged(object sender, EventArgs e) {
			m_Core.Config.ReadDelay = numReadDelay.Value / 1000m;
		}

		/// <summary>
		/// Raised when the step delay is changed.
		/// </summary>
		private void StepDelayChanged(object sender, EventArgs e) {
			m_Core.Config.MaxStepRate = Convert.ToUInt32(numMaxStepRate.Value);
		}

		/// <summary>
		/// Raised when the 'use timed pulse' context menu is selected (alternative for double-clicking).
		/// </summary>
		private void UseTimedPulseClick(object sender, EventArgs e) {
			var source = ((sender as ToolStripDropDownItem).Owner as ContextMenuStrip).SourceControl;
			if (source is Label)
				SetStepChecked(source as Label, false);
			else if (source is Control)
				SetStepChecked((source as Control).Tag as Label, false);
		}

		/// <summary>
		/// Raised when the 'use stepped pulse' context menu is selected (alternative for double-clicking).
		/// </summary>
		private void UseSteppedPulseClick(object sender, EventArgs e) {
            if (m_Core.Config.MotorType != MotorTypes.AutoCapV1) return;
			var source = ((sender as ToolStripDropDownItem).Owner as ContextMenuStrip).SourceControl;
			if (source is Label)
				SetStepChecked(source as Label, true);
			else if (source is Control)
				SetStepChecked((source as Control).Tag as Label, true);
		}

		/// <summary>
		/// Raised when the 'SWR Only' box is checked.
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void SwrOnlyChecked(object sender, EventArgs e) {
			m_Core.Config.ReadOnlySWR = cbReadOnlySWR.Checked;
		}

		/// <summary>
		/// Raised when the minimum power is changed.
		/// </summary>
		private void MinPowerChanged(object sender, EventArgs e) {
			m_Core.Config.MinPower = numMinPower.Value;
		}

		/// <summary>
		/// Rased when the "motor reversed" check is changed.
		/// </summary>
		private void MotorReverseChanged(object sender, EventArgs e) {
			m_Core.Config.MotorPolarityReversed = cbMotorReverse.Checked;
		}

		/// <summary>
		/// Raised when the relay roles (232 motor control) are reversed.
		/// </summary>
		private void MotorRelayRolesReversedClick(object sender, EventArgs e) {
			m_Core.Config.MotorRelayRolesReversed = !m_Core.Config.MotorRelayRolesReversed;
		}

		/// <summary>
		/// Raised when the motor output port is changed.
		/// </summary>
		private void MotorOutputPortChanged(object sender, EventArgs e) {
			m_Core.Config.MotorOutputPort = cbMotorOutputPort.SelectedIndex;
		}

		/// <summary>
		/// Raised when braking checkbox is changed.
		/// </summary>
		private void MotorBrakingChanged(object sender, EventArgs e) {
			m_Core.Config.MotorBrakingEnabled = cbMotorBraking.Checked;
		}

		/// <summary>
		/// Raised when the UP button is clicked.
		/// </summary>
		private void UpClick(object sender, EventArgs e) {
			if (IsStepChecked(lManualPulseMode)) {
				m_Core.Step(numManualEffort.Value / 100m, numManualPulse.Value);
			} else {
				m_Core.Pulse(numManualEffort.Value / 100m, numManualPulse.Value / 1000m);
			}
		}

		/// <summary>
		/// Raised when the DOWN button is clicked.
		/// </summary>
		private void DownClick(object sender, EventArgs e) {
			if (IsStepChecked(lManualPulseMode)) {
				m_Core.Step(-numManualEffort.Value / 100m, numManualPulse.Value);
			} else {
				m_Core.Pulse(-numManualEffort.Value / 100m, numManualPulse.Value / 1000m);
			}
		}

		/// <summary>
		/// One of the fast buttons was released.
		/// </summary>
		private void FastMouseUp(object sender, MouseEventArgs e) {
			if (m_LatchingMoves) return;
			m_Core.Move(0m);
		}

		/// <summary>
		/// One of the fast buttons was depressed.
		/// </summary>
		private void FastMouseDown(object sender, MouseEventArgs e) {
			if (m_LatchingMoves) return;
			m_Core.Move((sender == bFastUp ? numManualEffort.Value : -numManualEffort.Value) / 100m);
		}

		/// <summary>
		/// Raised when latch button is clicked.
		/// </summary>
		private void LatchClick(object sender, EventArgs e) {
			m_LatchingMoves = !m_LatchingMoves;
			if (m_LatchingMoves) {
				bLatchMovement.BackColor = Color.Lime;
			} else {
				bLatchMovement.BackColor = SystemColors.Control;
			}
		}

		/// <summary>
		/// Raised when the 'Radio Learning' checkbox changes.
		/// </summary>
		private void RadioLearningCheckedChanged(object sender, EventArgs e) {
			m_Core.Config.RadioLearningEnabled = cbRadioLearning.Checked;
		}

		/// <summary>
		/// Save a new profile.
		/// </summary>
		private void SaveProfileClick(object sender, EventArgs e) {
			using (var dialog = new SaveFileDialog {
					CheckPathExists = true,
					DefaultExt = ".xml",
					Filter = "Profile Files|*.xml|All Files|*.*",
					OverwritePrompt = true,
					AddExtension = true,
					InitialDirectory = Platform.UserProfileFolder,
					Title = "Create or Update Profile" }) {
				if (dialog.ShowDialog() != System.Windows.Forms.DialogResult.OK) {
					return;
				}
				SaveConfig(dialog.FileName);
			}
		}

		/// <summary>
		/// Load a new profile.
		/// </summary>
		private void LoadProfileClick(object sender, EventArgs e) {
			using (var dialog = new OpenFileDialog {
				CheckFileExists = true,
				CheckPathExists = true,
				DefaultExt = ".xml",
				Filter = "Profile Files|*.xml|All Files|*.*",
				AddExtension = true,
				InitialDirectory = Platform.UserProfileFolder,
				Multiselect = false,
				Title = "Load Profile",
			}) {
				if (dialog.ShowDialog() != System.Windows.Forms.DialogResult.OK) {
					return;
				}
				LoadConfig(dialog.FileName);
			}
		}

		/// <summary>
		/// Raised when 'About...' is clicked.
		/// </summary>
		private void AboutClick(object sender, EventArgs e) {
			MessageBox.Show(Platform.VersionString + "\r\n\r\n" +
				"Copyright (C) 2014-2016 by Matt Roberts, KK5JY, " +
				"All rights reserved.\r\n" +
				"\r\n" +
				"License: GNU General Public License Version 3.0\r\n" +
				"\r\n" +
				"This program is free software: you can redistribute it and/or modify " +
				"it under the terms of the GNU General Public License as published by " +
				"the Free Software Foundation, either version 3 of the License, or " +
				"(at your option) any later version.\r\n" +
				"\r\n" +
				"This program is distributed in the hope that it will be useful, but " +
				"WITHOUT ANY WARRANTY; without even the implied warranty of " +
				"MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU " +
				"General Public License for more details.\r\n" +
				"\r\n" +
				"You should have received a copy of the GNU General Public License " +
				"along with this program.  If not, see: http://www.gnu.org/licenses/",
				"About AutoCap...");
		}
		#endregion

		#region Utilities
		/// <summary>
		/// Test step label to see if it is 'checked'
		/// </summary>
		private static bool IsStepChecked(Label label) {
			return label.Text.Equals(PulseStepText, StringComparison.InvariantCultureIgnoreCase);
		}

		/// <summary>
		/// Set step label 'checked' state
		/// </summary>
		private void SetStepChecked(Label label, bool newValue) {
			if (ReferenceEquals(label, null)) return;
			label.Text = newValue ? PulseStepText : PulseTimeText;
			ValidateTimings();
		}

		/// <summary>
		/// Format the frequency.
		/// </summary>
		/// <param name="freqInHz">The frequency value in Hz.</param>
		/// <returns>A nicely formatted string in ham-speak form.</returns>
		private string FormatFrequency(decimal freqInHz) {
			var freq = Math.Truncate(freqInHz / 100m) / 10000m;
			var start = String.Format(CultureInfo.InvariantCulture, "{0:F4}", freq);
			var result = start.Substring(0, start.Length - 1);
			result += ".";
			result += start[start.Length - 1];
			result += " MHz";
			return result;
		}

		/// <summary>
		/// Load a configuration file.
		/// </summary>
		/// <remarks>
		/// This file is used both for startup/shutdown persistence,
		/// and for profile management.
		/// </remarks>
		/// <param name="path">Full path to the file.</param>
		private void LoadConfig(string path = null) {
			// load configuration
			bool complainOnFailure = true;
            try {
                var doc = new XmlDocument();
                if (String.IsNullOrEmpty(path)) {
                    path = Platform.ConfigFilePath;
                    complainOnFailure = false;
                }
                doc.Load(path);
                CoreConfig config = new CoreConfig();
                config.Deserialize(doc);

                // consume custom attributes from the root element
                if (doc.DocumentElement.HasAttribute("UserPulse")) {
                    numManualPulse.Value = Decimal.Parse(doc.DocumentElement.GetAttribute("UserPulse"), CultureInfo.InvariantCulture);
                }
                if (doc.DocumentElement.HasAttribute("UserEffort")) {
                    numManualEffort.Value = Decimal.Parse(doc.DocumentElement.GetAttribute("UserEffort"), CultureInfo.InvariantCulture);
                }
				if (doc.DocumentElement.HasAttribute("UserStep")) {
					SetStepChecked(lManualPulseMode, doc.DocumentElement.GetAttribute("UserStep").Equals("True", StringComparison.InvariantCultureIgnoreCase));
				}
                if (doc.DocumentElement.HasAttribute("FreqHint")) {
                    m_Core.FrequencyHint = Int32.Parse(doc.DocumentElement.GetAttribute("FreqHint"));
                }

                // consume the standard items
                numMinPower.Value = config.MinPower;
                cbMotorReverse.Checked = config.MotorPolarityReversed;
                cbMotorRelayReversed.Checked = !config.MotorRelayRolesReversed;
                cbMotorOutputPort.SelectedIndex = config.MotorOutputPort;
                cbMotorBraking.Checked = config.MotorBrakingEnabled;
                // FIXME: this next step should NOT be needed, but it's the only
                //        way to get ComboBox to commit the new Text value
                {
                    IPEndPoint ep;
                    if (PortUtils.TryParse(config.MotorControlPort, out ep)) {
                        cbMotorComPort.Items.Add(config.MotorControlPort);
                    }
                }
                cbMotorComPort.Text = config.MotorControlPort;
                switch (config.MotorType) {
                    case MotorTypes.RS232:
                        rbMotor232.Checked = true;
                        break;
                    case MotorTypes.RS232Bridge:
                        rbMotor232Bridge.Checked = true;
                        break;
                    case MotorTypes.AutoCapV1:
                        rbMotorAutoCapV1.Checked = true;
                        break;
#if DEBUG
                    case MotorTypes.Sim:
                        rbMotorSim.Checked = true;
                        break;
#endif
                }
                cbRadioHost.Text = config.RadioHost.ToString();
				cbRadioLearning.Checked = config.RadioLearningEnabled;
                switch (config.RadioType) {
                    case RadioTypes.FLDigi:
                        rbRadioFLDigi.Checked = true;
                        break;
                    case RadioTypes.HRD:
                        rbRadioHRD.Checked = true;
                        break;
                    case RadioTypes.None:
                        rbRadioNone.Checked = true;
                        break;
#if DEBUG
                    case RadioTypes.Sim:
                        rbRadioSim.Checked = true;
                        break;
#endif
                }
				cbReadOnlySWR.Checked = config.ReadOnlySWR;
				numReadDelay.Value = config.ReadDelay * 1000m;
				numMaxStepRate.Value = config.MaxStepRate;
                // FIXME: this next step should NOT be needed, but it's the only
                //        way to get ComboBox to commit the new Text value
                {
                    IPEndPoint ep;
                    if (PortUtils.TryParse(config.SensorPort, out ep)) {
                        cbSensorPort.Items.Add(config.SensorPort);
                    }
                }
                cbSensorPort.Text = config.SensorPort;
                switch (config.SensorType) {
                    case SensorTypes.AutoCapV1:
                        rbSensorAutoCapV1.Checked = true;
                        break;
                    case SensorTypes.LP100A:
                        rbSensorLP100A.Checked = true;
                        break;
#if DEBUG
                    case SensorTypes.Sim:
                        rbSensorSim.Checked = true;
                        break;
#endif
                }

                numSWR4.Value = config.Thresholds[0].SWR;
                numPulse4.Value = config.Thresholds[0].Size * (config.Thresholds[0].IsStep ? 1m : 1000m);
                numEffort4.Value = config.Thresholds[0].Effort * 100m;
				SetStepChecked(lStepUnit4, config.Thresholds[0].IsStep);

                numSWR3.Value = config.Thresholds[1].SWR;
				numPulse3.Value = config.Thresholds[1].Size * (config.Thresholds[1].IsStep ? 1m : 1000m);
                numEffort3.Value = config.Thresholds[1].Effort * 100m;
				SetStepChecked(lStepUnit3, config.Thresholds[1].IsStep);

                numSWR2.Value = config.Thresholds[2].SWR;
				numPulse2.Value = config.Thresholds[2].Size * (config.Thresholds[2].IsStep ? 1m : 1000m);
                numEffort2.Value = config.Thresholds[2].Effort * 100m;
				SetStepChecked(lStepUnit2, config.Thresholds[2].IsStep);

                numSWR1.Value = config.Thresholds[3].SWR;
				numPulse1.Value = config.Thresholds[3].Size * (config.Thresholds[3].IsStep ? 1m : 1000m);
                numEffort1.Value = config.Thresholds[3].Effort * 100m;
				SetStepChecked(lStepUnit1, config.Thresholds[3].IsStep);
			} catch {
				if (complainOnFailure) {
					MessageBox.Show(
						String.Format("Could not restore the saved profile from {0}", path),
						"Error"
					);
				}
				return;
			}
			Debug.WriteLine(String.Format("DEBUG: Loaded profile from: {0}", path));
		}

		/// <summary>
		/// Save a configuration file.
		/// </summary>
		/// <remarks>
		/// This file is used both for startup/shutdown persistence,
		/// and for profile management.
		/// </remarks>
		/// <param name="path">Full path to the file.</param>
		private void SaveConfig(string path = null) {
			// serialize the configuration
			var config = m_Core.Config.Serialize();
			var complainOnFailure = true;
			try {
				if (String.IsNullOrEmpty(path)) {
					path = Platform.ConfigFilePath;
					complainOnFailure = false;
				}
				var folder = System.IO.Path.GetDirectoryName(path);
				if (!System.IO.Directory.Exists(folder))
					System.IO.Directory.CreateDirectory(folder);
				// add custom attributes to the root element
				config.DocumentElement.SetAttribute("UserPulse", numManualPulse.Value.ToString(CultureInfo.InvariantCulture));
				config.DocumentElement.SetAttribute("UserEffort", numManualEffort.Value.ToString(CultureInfo.InvariantCulture));
				config.DocumentElement.SetAttribute("UserStep", IsStepChecked(lManualPulseMode).ToString(CultureInfo.InvariantCulture));
				config.DocumentElement.SetAttribute("FreqHint", m_Core.FrequencyHint.ToString(CultureInfo.InvariantCulture));
				config.Save(path);
			} catch {
				if (complainOnFailure) {
					MessageBox.Show(
						String.Format("Could not store the profile in {0}", path),
						"Error"
					);
				}
				return;
			}
			Debug.WriteLine(String.Format("DEBUG: Saved current profile to: {0}", path));
		}

		/// <summary>
		/// Commit the UI timings to the core.
		/// </summary>
		private void ValidateTimings() {
			var swr3ok = (numSWR4.Value > numSWR3.Value);
			var swr2ok = (numSWR3.Value > numSWR2.Value);
			var swr1ok = (numSWR2.Value > numSWR1.Value);
			numSWR3.BackColor = swr3ok ? SystemColors.Window : Color.Gold;
			numSWR2.BackColor = swr2ok ? SystemColors.Window : Color.Gold;
			numSWR1.BackColor = swr1ok ? SystemColors.Window : Color.Gold;
			var step4 = IsStepChecked(lStepUnit4);
			var step3 = IsStepChecked(lStepUnit3);
			var step2 = IsStepChecked(lStepUnit2);
			var step1 = IsStepChecked(lStepUnit1);
			numMaxStepRate.Enabled = (step4 || step3 || step2 || step1);
#if false	// for now, stepper read delay requires both settings, so leave this one enabled
			numReadDelay.Enabled = !(step4 && step3 && step2 && step1);
#endif

			// if all settings OK, commit to core config
			if (swr1ok && swr2ok && swr3ok) {
				m_Core.Config.Thresholds = new[] {
					new Threshold { SWR = numSWR4.Value, Size = numPulse4.Value / (step4 ? 1m : 1000m), Effort = numEffort4.Value / 100m, IsStep = step4 },
					new Threshold { SWR = numSWR3.Value, Size = numPulse3.Value / (step3 ? 1m : 1000m), Effort = numEffort3.Value / 100m, IsStep = step3 },
					new Threshold { SWR = numSWR2.Value, Size = numPulse2.Value / (step2 ? 1m : 1000m), Effort = numEffort2.Value / 100m, IsStep = step2 },
					new Threshold { SWR = numSWR1.Value, Size = numPulse1.Value / (step1 ? 1m : 1000m), Effort = numEffort1.Value / 100m, IsStep = step1 },
				};
			}
		}

		/// <summary>
		/// Test a host name for validity.
		/// </summary>
		private void ValidateRadioSourceAddress(string name) {
			IPAddress ip = null;
			if (!IPAddress.TryParse(name, out ip)) {
				ip = null;
			}
			if (ip == null) {
				try {
					// grab the first IP address offered
					ip = Dns.GetHostEntry(name).AddressList.FirstOrDefault();
				} catch {
					// nop
				}
			}
			if (ip == null) {
				m_LastRadioHostValid = false;
				return;
			}
			m_LastRadioHostValid = true;
			m_LastRadioAddress = name;
		}

		/// <summary>
		/// Test serial port names for validity.
		/// </summary>
		private void ValidatePorts() {
            IPEndPoint ep;

			// check motor control port
			if (m_Core.Config.MotorType == MotorTypes.Sim) {
				m_Messages["MotorPort"] = "";
			} else {
                if (m_PortNames.Contains(cbMotorComPort.Text.Replace(":", "")) || PortUtils.TryParse(cbMotorComPort.Text, out ep)) {
					m_Core.Config.MotorControlPort = cbMotorComPort.Text;
					m_Messages["MotorPort"] = "";
				} else {
					m_Messages["MotorPort"] = "Motor Control Port Invalid";
				}
			}

			// check sensor port
			if (m_Core.Config.SensorType == SensorTypes.Sim) {
				m_Messages["SensorPort"] = "";
			} else {
                if (m_PortNames.Contains(cbSensorPort.Text.Replace(":", "")) || PortUtils.TryParse(cbSensorPort.Text, out ep)) {
					m_Core.Config.SensorPort = cbSensorPort.Text;
					m_Messages["SensorPort"] = "";
				} else {
					m_Messages["SensorPort"] = "Sensor Port Invalid";
				}
			}

			// display immediately
			m_MessageTimer = 0;
		}
		#endregion
	}
}
